"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.compileClauseMatrix = exports.isNeeded = exports.isRowUseless = exports.isPatternRedundant = exports.isPatternMoreGeneral = exports.heuristicConstructorPrefix = exports.heuristicArity = exports.heuristicSmallBranchingFactor = exports.heuristicNeededPrefix = exports.selectColumn = exports.defaultClauseMatrix = exports.specializeClauseMatrix = exports.clauseMatrixOf = exports.occTermOfRule = exports.substituteOccurences = exports.occurencesOf = exports.patternOf = exports._ = void 0;
const Types_1 = require("../../Parser/Types");
const Utils_1 = require("../../Parser/Utils");
const Utils_2 = require("../Utils");
const DecisionTree_1 = require("./DecisionTree");
exports._ = '_';
const patternOf = (term) => {
    if (Utils_2.isVar(term))
        return exports._;
    return { name: term.name, args: term.args.map(exports.patternOf) };
};
exports.patternOf = patternOf;
const occurencesOf = (term, sigma = {}, subTermIndex = 0) => {
    if (Utils_2.isVar(term))
        return Types_1.dictSet(sigma, term, { index: subTermIndex, pos: [] });
    const collectOccurences = (t, sigma, localOffset = 0, parent) => {
        const occ = {
            index: parent.index,
            pos: [...parent.pos, localOffset]
        };
        if (Utils_2.isVar(t)) {
            sigma[t] = occ;
            return;
        }
        t.args.forEach((s, idx) => {
            collectOccurences(s, sigma, idx, occ);
        });
    };
    for (const [t, i] of Utils_1.indexed(term.args)) {
        collectOccurences(t, sigma, i, { index: subTermIndex, pos: [] });
    }
    return sigma;
};
exports.occurencesOf = occurencesOf;
const substituteOccurences = (term, occs) => {
    if (Utils_2.isVar(term)) {
        if (Types_1.dictHas(occs, term)) {
            return Types_1.dictGet(occs, term);
        }
        else {
            throw new Error(`unbound variable: ${term} in ${JSON.stringify(occs)}`);
        }
    }
    return {
        name: term.name,
        args: term.args.map(arg => exports.substituteOccurences(arg, occs))
    };
};
exports.substituteOccurences = substituteOccurences;
const occTermOfRule = ([lhs, rhs]) => {
    const sigma = {};
    for (const [arg, i] of Utils_1.indexed(lhs.args)) {
        exports.occurencesOf(arg, sigma, i);
    }
    return exports.substituteOccurences(rhs, sigma);
};
exports.occTermOfRule = occTermOfRule;
// all the rules must share the same head symbol and arity
const clauseMatrixOf = (rules) => {
    return {
        patterns: rules.map(rule => Utils_2.lhs(rule).args.map(exports.patternOf)),
        dims: [rules.length, Utils_2.arity(rules)],
        actions: rules.map(exports.occTermOfRule)
    };
};
exports.clauseMatrixOf = clauseMatrixOf;
const specializeRow = (row, ctor, arity) => {
    const [p, ps] = Utils_2.decons(row);
    if (p === exports._)
        return [...Utils_1.repeat(exports._, arity), ...ps];
    if (p.name === ctor)
        return [...p.args, ...ps];
};
const specializeClauseMatrix = (matrix, c, arity) => {
    const patterns = matrix.patterns
        .map(row => specializeRow(row, c, arity));
    const actions = [...Utils_2.zip(patterns, matrix.actions)]
        .filter(([p, _a]) => p)
        .map(([_p, a]) => a);
    return {
        patterns: patterns.filter(Utils_2.isSomething),
        dims: [actions.length, arity + matrix.dims[1] - 1],
        actions
    };
};
exports.specializeClauseMatrix = specializeClauseMatrix;
const defaultRow = (row) => {
    const [p, ps] = Utils_2.decons(row);
    if (p === exports._)
        return ps;
};
const defaultClauseMatrix = (matrix) => {
    const patterns = matrix.patterns.map(defaultRow);
    const actions = [...Utils_2.zip(patterns, matrix.actions)]
        .filter(([p, _a]) => p)
        .map(([_p, a]) => a);
    return {
        patterns: patterns.filter(Utils_2.isSomething),
        dims: [actions.length, matrix.dims[1] - 1],
        actions
    };
};
exports.defaultClauseMatrix = defaultClauseMatrix;
const getColumn = (matrix, i) => {
    const col = [];
    for (const row of matrix.patterns) {
        col.push(row[i]);
    }
    return col;
};
const selectColumn = (matrix) => {
    for (let i = 0; i < matrix.dims[1]; i++) {
        if (getColumn(matrix, i).some(p => p !== exports._)) {
            return i;
        }
    }
    Types_1.unreachable('No valid column found');
};
exports.selectColumn = selectColumn;
const bestCandidates = (...scores) => {
    const maxScore = Math.max(...scores);
    const indices = [];
    for (const [score, idx] of Utils_1.indexed(scores)) {
        if (score === maxScore) {
            indices.push(idx);
        }
    }
    return indices;
};
// p
const heuristicNeededPrefix = (matrix) => {
    const score = (col) => {
        const allNeeded = (upToRow) => {
            for (let row = 0; row <= upToRow; row++) {
                if (!exports.isNeeded(row, col, matrix))
                    return false;
            }
            return true;
        };
        for (let row = matrix.dims[0] - 1; row >= 0; row--) {
            if (allNeeded(row))
                return row;
        }
        return 0;
    };
    return bestCandidates(...Utils_1.map(Utils_1.range(0, matrix.dims[1] - 1), col => score(col)));
};
exports.heuristicNeededPrefix = heuristicNeededPrefix;
// b
const heuristicSmallBranchingFactor = (matrix) => {
    const score = (col) => {
        return -heads(getColumn(matrix, col)).size - 1;
    };
    return bestCandidates(...Utils_1.map(Utils_1.range(0, matrix.dims[1] - 1), col => score(col)));
};
exports.heuristicSmallBranchingFactor = heuristicSmallBranchingFactor;
// a
const heuristicArity = (matrix) => {
    const score = (col) => {
        const headSymbols = new Set();
        let sum = 0;
        for (const pat of getColumn(matrix, col)) {
            if (pat !== exports._ && !headSymbols.has(pat.name)) {
                sum += pat.args.length;
                headSymbols.add(pat.name);
            }
        }
        return -sum;
    };
    return bestCandidates(...Utils_1.map(Utils_1.range(0, matrix.dims[1] - 1), col => score(col)));
};
exports.heuristicArity = heuristicArity;
const heuristicConstructorPrefix = (matrix) => {
    const score = (col) => {
        const allNeeded = (upToRow) => {
            for (let row = 0; row <= upToRow; row++) {
                if (matrix.patterns[row][col] === exports._)
                    return false;
            }
            return true;
        };
        for (let row = matrix.dims[0] - 1; row >= 0; row--) {
            if (allNeeded(row))
                return row;
        }
        return 0;
    };
    return bestCandidates(...Utils_1.map(Utils_1.range(0, matrix.dims[1] - 1), col => score(col)));
};
exports.heuristicConstructorPrefix = heuristicConstructorPrefix;
const combineHeuristics = (heuristics) => {
    return (matrix) => {
        const selected = new Set(Utils_1.range(0, matrix.dims[1] - 1));
        for (const heuristic of heuristics) {
            if (selected.size === 1)
                break;
            Utils_1.intersectionMut(selected, new Set(heuristic(matrix)));
        }
        // return the first suitable column
        return selected.values().next().value;
    };
};
const isPatternMoreGeneral = (p1, p2) => {
    if (p1 === exports._)
        return true;
    if (p2 === exports._)
        return false;
    if (p1.name !== p2.name)
        return false;
    return Utils_1.every(Utils_2.zip(p1.args, p2.args), ([a, b]) => exports.isPatternMoreGeneral(a, b));
};
exports.isPatternMoreGeneral = isPatternMoreGeneral;
// see "The Definition of Standard ML", Milner et al. 1990
const isPatternRedundant = (patternIndex, patterns) => {
    const p = patterns[patternIndex];
    for (let i = 0; i < patternIndex; i++) {
        if (exports.isPatternMoreGeneral(patterns[i], p)) {
            return true;
        }
    }
    return false;
};
exports.isPatternRedundant = isPatternRedundant;
const isRowUseless = (matrix, rowIndex) => {
    for (let col = 0; col < matrix.dims[1]; col++) {
        if (!exports.isPatternRedundant(rowIndex, getColumn(matrix, col))) {
            return false;
        }
    }
    return true;
};
exports.isRowUseless = isRowUseless;
// P/ω
const filterColumn = (matrix, colIndex) => {
    const [rows, cols] = matrix.dims;
    const filtered = matrix.patterns.map(row => row.filter((_, i) => i !== colIndex));
    return {
        dims: [rows, cols - 1],
        actions: matrix.actions,
        patterns: filtered
    };
};
const isNeeded = (row, col, matrix) => {
    if (matrix.patterns[row][col] !== exports._)
        return true;
    return exports.isRowUseless(filterColumn(matrix, col), row);
};
exports.isNeeded = isNeeded;
const swapColumn = (matrix, i) => {
    for (const row of matrix.patterns) {
        Utils_2.swapMut(row, 0, i);
    }
};
const heads = (patterns) => {
    const hds = new Map();
    for (const p of patterns) {
        if (p !== exports._) {
            hds.set(p.name, p.args.length);
        }
    }
    return hds;
};
const compileClauseMatrix = (argsCount, matrix, signature, heuristics = [
    exports.heuristicNeededPrefix,
    exports.heuristicSmallBranchingFactor,
    exports.heuristicArity // a
]) => {
    const occurences = [...Utils_1.gen(argsCount, i => ({ index: i, pos: [] }))];
    const selectColumn = combineHeuristics(heuristics);
    return compileClauseMatrixAux(occurences, matrix, signature, selectColumn);
};
exports.compileClauseMatrix = compileClauseMatrix;
const compileClauseMatrixAux = (occurences, matrix, signature, selectColumn) => {
    var _b;
    const [m, n] = matrix.dims;
    if (m === 0)
        return DecisionTree_1.makeFail();
    if (m > 0 && (n === 0 || matrix.patterns[0].every(p => p === exports._))) {
        return DecisionTree_1.makeLeaf(matrix.actions[0]);
    }
    const colIdx = (_b = selectColumn(matrix)) !== null && _b !== void 0 ? _b : 0;
    if (colIdx !== 0) {
        Utils_2.swapMut(occurences, 0, colIdx);
        swapColumn(matrix, colIdx);
    }
    const col = getColumn(matrix, 0);
    const hds = heads(col);
    const tests = [];
    for (const [ctor, arity] of hds) {
        const o1 = [...Utils_1.gen(arity, i => ({
                index: occurences[0].index,
                pos: [...occurences[0].pos, i]
            }))];
        const A_k = compileClauseMatrixAux([...o1, ...Utils_2.tail(occurences)], exports.specializeClauseMatrix(matrix, ctor, arity), signature, selectColumn);
        tests.push([ctor, A_k]);
    }
    if (!Utils_2.setEq(hds, signature)) {
        const A_D = compileClauseMatrixAux(Utils_2.tail(occurences), exports.defaultClauseMatrix(matrix), signature, selectColumn);
        tests.push([exports._, A_D]);
    }
    return DecisionTree_1.makeSwitch(Utils_2.head(occurences), tests);
};
