"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.collectArity = exports.collectTRSArities = exports.lazify = exports.Thunk = exports.Inst = exports.MaybeLazy = exports.Lazy = exports.thunkSymb = exports.instSymb = exports.lazyAnnotationSymb = void 0;
const Types_1 = require("../../Types");
const Utils_1 = require("../Utils");
exports.lazyAnnotationSymb = 'Lazy';
exports.instSymb = 'inst';
exports.thunkSymb = 'Θ';
const Lazy = (term) => Utils_1.fun(exports.lazyAnnotationSymb, term);
exports.Lazy = Lazy;
const MaybeLazy = (term, lazy = true) => lazy ? Utils_1.fun(exports.lazyAnnotationSymb, term) : term;
exports.MaybeLazy = MaybeLazy;
const Inst = (...args) => Utils_1.fun(exports.instSymb, ...args);
exports.Inst = Inst;
const Thunk = (name, ...args) => Utils_1.fun(`${exports.thunkSymb}${name}`, ...args);
exports.Thunk = Thunk;
const lazify = (trsWithAnnotations) => {
    const [ann, hasLazyTerms] = collectLazinessAnnotations(trsWithAnnotations);
    const trs = removeLazinessAnnotations(trsWithAnnotations);
    if (!hasLazyTerms)
        return Types_1.Ok(trs);
    const newTrs = Utils_1.emptyTRS();
    const arities = exports.collectTRSArities(trs);
    // thunk each lazy argument
    for (const [name, rules] of trs.entries()) {
        for (const [lhs, rhs] of rules) {
            const thunkedRule = instantiateMigrants([
                thunkify(lhs, ann),
                thunkify(rhs, ann)
            ], ann);
            Utils_1.addRules(newTrs, thunkedRule);
            // make sure the arity is correct for function symbols
            arities.set(name, Utils_1.arity(rules));
        }
    }
    // instantiate each function symbol
    for (const [symb, ar] of arities.entries()) {
        if (ar > 0) {
            const varNames = Utils_1.genVars(ar);
            const instRule = [
                exports.Inst(exports.Thunk(symb, ...varNames)),
                instantiateLazyArgs(Utils_1.fun(symb, ...varNames), ann)
            ];
            Utils_1.addRules(newTrs, instRule);
        }
    }
    Utils_1.addRules(newTrs, [exports.Inst('x'), 'x']);
    return Types_1.Ok(newTrs);
};
exports.lazify = lazify;
const removeLazinessAnnotation = (term) => {
    if (Utils_1.isVar(term))
        return term;
    const newArgs = [];
    for (const t of (term.args).map(t => removeLazinessAnnotation(t))) {
        if (Utils_1.isFun(t, exports.lazyAnnotationSymb)) {
            newArgs.push(t.args[0]);
        }
        else {
            newArgs.push(t);
        }
    }
    return { name: term.name, args: newArgs };
};
const removeLazinessAnnotations = (trs) => {
    const newTrs = new Map();
    for (const rules of trs.values()) {
        for (const [lhs, rhs] of rules) {
            Utils_1.addRules(newTrs, [removeLazinessAnnotation(lhs), removeLazinessAnnotation(rhs)]);
        }
    }
    return newTrs;
};
const collectLazinessAnnotation = (term, annotations) => {
    var _a;
    if (Utils_1.isVar(term))
        return false;
    let hasLazyTerms = false;
    const name = term.name;
    const ar = term.args.length;
    const ann = (_a = annotations.get(name)) !== null && _a !== void 0 ? _a : Utils_1.fill(false, ar);
    term.args.forEach((t, i) => {
        if (Utils_1.isFun(t, exports.lazyAnnotationSymb)) {
            ann[i] = true;
            hasLazyTerms = true;
        }
        annotations.set(name, ann);
        hasLazyTerms = collectLazinessAnnotation(t, annotations) || hasLazyTerms;
    });
    return hasLazyTerms;
};
const collectLazinessAnnotations = (trs) => {
    const annotations = new Map();
    let hasLazyTerms = false;
    for (const rules of trs.values()) {
        for (const [lhs, rhs] of rules) {
            hasLazyTerms = collectLazinessAnnotation(lhs, annotations) || hasLazyTerms;
            hasLazyTerms = collectLazinessAnnotation(rhs, annotations) || hasLazyTerms;
        }
    }
    return [annotations, hasLazyTerms];
};
const collectTRSArities = (trs) => {
    const arities = new Map();
    for (const rules of trs.values()) {
        for (const rule of rules) {
            collectRuleArity(rule, arities);
        }
    }
    return arities;
};
exports.collectTRSArities = collectTRSArities;
const collectRuleArity = ([lhs, rhs], arities) => {
    exports.collectArity(lhs, arities);
    exports.collectArity(rhs, arities);
};
const collectArity = (t, arities) => {
    var _a;
    if (Utils_1.isFun(t)) {
        arities.set(t.name, (_a = arities.get(t.name)) !== null && _a !== void 0 ? _a : t.args.length);
        t.args.forEach(s => exports.collectArity(s, arities));
    }
};
exports.collectArity = collectArity;
const isLazy = (ruleName, argIdx, ann) => {
    return ann.has(ruleName) ? ann.get(ruleName)[argIdx] : false;
};
const isVarActive = (x, t, ann) => {
    if (Utils_1.isVar(t))
        return t === x;
    return t.args.some((arg, i) => (!isLazy(t.name, i, ann) && isVarActive(x, arg, ann)));
};
const isVarMigrant = (x, [lhs, rhs], ann) => {
    return !isVarActive(x, lhs, ann) && isVarActive(x, rhs, ann);
};
const migrantVars = (rule, ann) => {
    return Utils_1.ruleVars(rule).filter(x => isVarMigrant(x, rule, ann));
};
const thunkifyAll = (ts, ann) => (ts.map(t => (Utils_1.isVar(t) || Utils_1.isConst(t)) ? t : exports.Thunk(t.name, ...thunkifyAll(t.args, ann))));
function thunkify(t, ann) {
    if (Utils_1.isVar(t) || Utils_1.isConst(t))
        return t;
    const name = t.name;
    const args = t.args.map((arg, i) => {
        if (Utils_1.isFun(arg) && isLazy(name, i, ann) && !Utils_1.isConst(arg)) {
            return thunkify(exports.Thunk(arg.name, ...thunkifyAll(arg.args, ann)), ann);
        }
        return thunkify(arg, ann);
    });
    return { name, args };
}
;
const instantiateMigrants = (rule, ann) => {
    const [l, r] = rule;
    if (Utils_1.isVar(r) && isVarMigrant(r, rule, ann)) {
        return [l, exports.Inst(r)];
    }
    else if (Utils_1.isFun(r)) {
        const migrants = migrantVars(rule, ann);
        if (!Utils_1.isEmpty(migrants)) {
            const newArgs = migrants.reduce((ts, x) => Utils_1.replaceTerms(x, exports.Inst(x), ts), r.args);
            const newRhs = { name: r.name, args: newArgs };
            return [l, newRhs];
        }
    }
    return rule;
};
const instantiateLazyArgs = (term, annotations) => {
    if (Utils_1.isVar(term))
        return term;
    const newArgs = [];
    if (!annotations.has(term.name))
        return term;
    const ann = annotations.get(term.name);
    term.args.map(t => instantiateLazyArgs(t, annotations)).forEach((t, i) => {
        if (!ann[i]) {
            newArgs.push(exports.Inst(t));
        }
        else {
            newArgs.push(t);
        }
    });
    return Utils_1.fun(term.name, ...newArgs);
};
