"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.arithmeticExternals = exports.symb = void 0;
const Imports_1 = require("../Compiler/Passes/Imports");
const Utils_1 = require("../Compiler/Utils");
const Translator_1 = require("../Translator/Translator");
const symb = (f) => ({ name: f, args: [] });
exports.symb = symb;
const arithbinop = (t, op) => {
    const [a, b] = t.args;
    if (Utils_1.isFun(a) && Utils_1.isFun(b)) {
        try {
            const an = BigInt(a.name);
            const bn = BigInt(b.name);
            const res = op(an, bn);
            return exports.symb(`${res}`);
        }
        catch (e) {
            return t;
        }
    }
    return t;
};
const boolbinop = (t, op) => {
    const [a, b] = t.args;
    if (Utils_1.isFun(a) && Utils_1.isFun(b)) {
        const an = BigInt(a.name);
        const bn = BigInt(b.name);
        const res = op(an, bn);
        return res ? Imports_1.True : Imports_1.False;
    }
    return t;
};
const nativeArithmeticExternals = {
    'add': t => arithbinop(t, (a, b) => a + b),
    'sub': t => arithbinop(t, (a, b) => a - b),
    'mult': t => arithbinop(t, (a, b) => a * b),
    'div': t => arithbinop(t, (a, b) => a / b),
    'mod': t => arithbinop(t, (a, b) => a % b),
    'gtr': t => boolbinop(t, (a, b) => a > b),
    'geq': t => boolbinop(t, (a, b) => a >= b),
    'lss': t => boolbinop(t, (a, b) => a < b),
    'leq': t => boolbinop(t, (a, b) => a <= b)
};
const jsarithbinop = (op) => {
    return (name) => (`function ${name}(a, b) {
            if (isNat(a) && isNat(b)) {
                return ${op === '/' ? 'typeof a === "number" ? Math.floor(a / b) : a / b' : `a ${op} b`};
            }
        
            return { name: "${name}", args: [a, b] };
        }`);
};
const jsboolbinop = (op) => {
    return (name) => (`function ${name}(a, b) {
            if (isNat(a) && isNat(b)) {
                return (a ${op} b) ? ${Translator_1.nullaryVarName('True')} : ${Translator_1.nullaryVarName('False')};
            }
        
            return { name: "${name}", args: [a, b] };
        }`);
};
const jsArithmeticExternals = {
    'add': jsarithbinop('+'),
    'sub': jsarithbinop('-'),
    'mult': jsarithbinop('*'),
    'div': jsarithbinop('/'),
    'mod': jsarithbinop('%'),
    'gtr': jsboolbinop('>'),
    'geq': jsboolbinop('>='),
    'lss': jsboolbinop('<'),
    'leq': jsboolbinop('<=')
};
const haskellIntBinop = (op) => (name) => `${name} (Nat a) (Nat b) = Nat (a ${op} b)`;
const haskellBoolBinop = (op) => (name) => `${name} (Nat a) (Nat b) =
        if a ${op} b then ${Translator_1.nullaryVarName('True')} else ${Translator_1.nullaryVarName('False')}`;
const haskellArithmeticExternals = {
    'add': haskellIntBinop('+'),
    'sub': haskellIntBinop('-'),
    'mult': haskellIntBinop('*'),
    'div': haskellIntBinop('`div`'),
    'mod': haskellIntBinop('`mod`'),
    'gtr': haskellBoolBinop('>'),
    'geq': haskellBoolBinop('>='),
    'lss': haskellBoolBinop('<'),
    'leq': haskellBoolBinop('<=')
};
const ocamlIntBinop = (op) => (name) => `let ${name} (a, b) = match (a, b) with
        | (Nat a, Nat b) -> Nat (a ${op} b)
        | _ -> Fun ("${name}", [a; b]);;
    `;
const ocamlBoolOf = (expr) => `(if ${expr} then ${Translator_1.nullaryVarName('True')} else ${Translator_1.nullaryVarName('False')})`;
const ocamlBoolBinop = (op) => (name) => `let ${name} (a, b) = match (a, b) with
        | (Nat a, Nat b) -> ${ocamlBoolOf(`a ${op} b`)}
        | _ -> Fun ("${name}", [a; b]);;
    `;
const ocamlArithmeticExternals = {
    "sub": ocamlIntBinop('-'),
    "add": ocamlIntBinop('+'),
    "mult": ocamlIntBinop('*'),
    "mod": ocamlIntBinop('mod'),
    "div": ocamlIntBinop('/'),
    "lss": ocamlBoolBinop('<'),
    "leq": ocamlBoolBinop('<='),
    "gtr": ocamlBoolBinop('>'),
    "geq": ocamlBoolBinop('>=')
};
const arithmeticExternals = (target) => {
    return {
        'native': nativeArithmeticExternals,
        'js': jsArithmeticExternals,
        'ocaml': ocamlArithmeticExternals,
        'haskell': haskellArithmeticExternals
    }[target];
};
exports.arithmeticExternals = arithmeticExternals;
