"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.metaExternals = void 0;
const Imports_1 = require("../Compiler/Passes/Imports");
const Lazify_1 = require("../Compiler/Passes/Lazify");
const Utils_1 = require("../Compiler/Utils");
const Normalizer_1 = require("../Normalizer/Normalizer");
const Utils_2 = require("../Parser/Utils");
const Translator_1 = require("../Translator/Translator");
const trace = (log) => (query, normalizer, externals) => {
    return Normalizer_1.traceNormalize(Lazify_1.Inst(query.args[0]), normalizer, externals, (t) => {
        log(Utils_1.showTerm(t));
    });
};
const equ = (s, t) => {
    return Utils_1.termsEq(s, t) ? Imports_1.True : Imports_1.False;
};
const nativeMetaExternals = (log) => {
    return {
        trace: trace(log),
        equ: t => { const [a, b] = t.args; return equ(a, b); },
        show: t => Utils_1.showTerm(t)
    };
};
const traceNotAvailableMsg = '@trace is not available with this normalizer';
const jsMetaExternals = {
    trace: name => `function ${name}() { throw new Error("${traceNotAvailableMsg}"); }`,
    equ: name => `
        function ${name}(a, b) {
            if (isFun(a) && isFun(b) && a.name === b.name && a.args.length === b.args.length) {
              for (let i = 0; i < a.args.length; i++) {
                  if (${name}(a.args[i], b.args[i]).name === "False") {
                      return ${Translator_1.nullaryVarName('False')};
                  }
              }
              return ${Translator_1.nullaryVarName('True')};
            }
          
            return a === b ? ${Translator_1.nullaryVarName('True')} : ${Translator_1.nullaryVarName('False')};
        }`,
    show: name => `
        function ${name}(term) {
            if (isVar(term)) return term;
            if (isNat(term)) return term.toString();
            if (term.args.length === 0) return term.name;
            
            const terms = [term];
            const stack = [];
            const symbols = [];
            
            // flatten the terms onto a stack
            while (terms.length > 0) {
                const t = terms.pop();
            
                if (isVar(t) || isNat(t)) {
                    stack.push(t.toString());
                } else if (t.args.length === 0) {
                    stack.push(t.name);
                } else {
                    for (let i = t.args.length - 1; i >= 0; i--) {
                        terms.push(t.args[i]);
                    }
                
                    terms.push(t.name);
                    symbols.push([t.name, t.args.length]);
                }
            }
            
            const argsStack = [];
            
            // assemble constructors back when all arguments have been stringified
            for (let i = stack.length - 1; i >= 0; i--) {
                const t = stack[i];
                if (symbols.length === 0) break;
                const [f, ar] = symbols[symbols.length - 1];
            
                if (t === f) {
                    const args = [];
                    for (let k = 0; k < ar; k++) {
                        args.push(argsStack.pop());
                    }
            
                    argsStack.push(f + '(' + args.join(', ') + ')');
                    symbols.pop();
                } else {
                    argsStack.push(t);
                }
            }
            
            return argsStack[0];
        }`
};
const haskellMetaExternals = {
    trace: name => `${name} = error "${traceNotAvailableMsg}"`,
    equ: name => `${name} a b = if a == b then ${Translator_1.nullaryVarName('True')} else ${Translator_1.nullaryVarName('False')}`,
    show: name => Utils_2.format(`${name} (Var x) = x`, `${name} (Fun f []) = f`, `${name} (Fun f ts) = f ++ "(" ++ (intercalate ", " (map ${name} ts)) ++ ")"`, `${name} (Nat n) = show n`)
};
const ocamlMetaExternals = {
    trace: name => `let ${name} () = failwith "${traceNotAvailableMsg}";;`,
    equ: name => `let ${name} (a, b) = if a = b then ${Translator_1.nullaryVarName('True')} else ${Translator_1.nullaryVarName('False')};;`,
    show: name => `let rec ${name} t = match t with
        | Var x -> x
        | Fun (f, []) -> f
        | Fun (f, ts) -> f ^ "(" ^ (String.concat ", " (List.map ${name} ts)) ^ ")"
        | Nat n -> string_of_int n;;
`
};
const log = (msg) => { console.log(msg); };
const metaExternals = (nativeLog = log) => (target) => {
    return {
        'native': nativeMetaExternals(nativeLog),
        'js': jsMetaExternals,
        'ocaml': ocamlMetaExternals,
        'haskell': haskellMetaExternals
    }[target];
};
exports.metaExternals = metaExternals;
