"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.buildMatcher = exports.unstringify = exports.symbs = exports.makePatterns = exports.termToNames = exports.genSymbolSplitter = exports.collectSymbols = exports.collectTermSymbols = exports.collectArities = exports.stringify = exports.closure = exports.prepend = exports.factorOut = exports.ω = exports.ε = void 0;
const Unification_1 = require("../../Unification");
const Parser_1 = require("../../../Parser/Parser");
const Utils_1 = require("../../../Parser/Utils");
const Lazify_1 = require("../../../Compiler/Passes/Lazify");
const Utils_2 = require("../../../Compiler/Utils");
const StringClosureMatcher_1 = require("./StringClosureMatcher");
exports.ε = '';
exports.ω = 'ω';
const factorOut = (α, M) => {
    const factored = new Set();
    for (const suffix of M) {
        if (suffix.substr(0, α.length) === α) {
            factored.add(suffix.substr(α.length));
        }
    }
    return factored;
};
exports.factorOut = factorOut;
const prepend = (α, M) => (Utils_1.setMap(M, s => `${α}${s}`));
exports.prepend = prepend;
const closure = (M, arities) => {
    if ((M.size === 1 && M.has(exports.ε)) || Utils_2.isEmpty(M))
        return M;
    const Mα = (α) => {
        const M_α = exports.factorOut(α, M);
        if (α === exports.ω)
            return M_α;
        if (Utils_2.isEmpty(M_α))
            return new Set();
        const arity = arities.get(α);
        return Utils_1.unionMut(M_α, exports.prepend(Utils_1.repeatString(exports.ω, arity), exports.factorOut(exports.ω, M)));
    };
    const M_ = new Set();
    for (const α of Utils_1.join(arities.keys(), Utils_1.once(exports.ω))) {
        Utils_1.unionMut(M_, exports.prepend(α, exports.closure(Mα(α), arities)));
    }
    return M_;
};
exports.closure = closure;
const stringify = (t) => {
    if (Utils_2.isNothing(t))
        return "";
    if (Utils_2.isVar(t))
        return exports.ω;
    return `${t.name}${t.args.map(s => exports.stringify(s)).join("")}`;
};
exports.stringify = stringify;
const collectArities = (terms) => {
    const arities = new Map();
    for (const t of terms) {
        if (t) {
            Lazify_1.collectArity(t, arities);
        }
    }
    return arities;
};
exports.collectArities = collectArities;
const collectTermSymbols = (t, symbols) => {
    if (Utils_2.isFun(t)) {
        symbols.add(t.name);
        t.args.forEach(s => exports.collectTermSymbols(s, symbols));
    }
};
exports.collectTermSymbols = collectTermSymbols;
const collectSymbols = (terms) => {
    const symbols = new Set();
    for (const t of terms) {
        if (t) {
            exports.collectTermSymbols(t, symbols);
        }
    }
    return symbols;
};
exports.collectSymbols = collectSymbols;
function genSymbolSplitter(symbols) {
    const maxLen = Math.max(...symbols.map(l => l.length));
    const alphabetSet = new Set(symbols);
    return (str) => {
        const letters = [];
        while (str.length > 0) {
            let found = false;
            for (let len = maxLen; len > 0 && str.length > 0; len--) {
                const prefix = str.substr(0, len);
                if (alphabetSet.has(prefix)) {
                    found = true;
                    letters.push(prefix);
                    str = str.slice(len);
                    break;
                }
            }
            if (!found)
                return;
        }
        return letters;
    };
}
exports.genSymbolSplitter = genSymbolSplitter;
const replaceVars = (t) => {
    if (Utils_2.isVar(t))
        return exports.ω;
    return Utils_2.fun(t.name, ...t.args.map(s => replaceVars(s)));
};
const termToNames = (term) => {
    return [...Utils_1.traverseNames(term)];
};
exports.termToNames = termToNames;
const makePatterns = (...patterns) => {
    return patterns
        .map(s => Parser_1.parseTerm(s))
        .filter(t => Utils_2.isSomething(t) && Utils_2.isFun(t))
        .map(replaceVars);
};
exports.makePatterns = makePatterns;
const symbs = (term) => {
    const [t] = exports.makePatterns(term);
    return exports.termToNames(t);
};
exports.symbs = symbs;
const unstringify = (str, splitter, arities) => {
    if (str === exports.ω)
        return exports.ω;
    const symbols = splitter(str);
    if (Utils_2.isSomething(symbols)) {
        return unstringifyAux(symbols, arities);
    }
};
exports.unstringify = unstringify;
const unstringifyAux = (symbols, arities) => {
    var _a;
    if (Utils_2.isEmpty(symbols))
        return;
    const stack = [];
    for (const f of Utils_1.reverse(symbols)) {
        if (f === exports.ω) {
            stack.push(exports.ω);
        }
        else {
            const ar = (_a = arities.get(f)) !== null && _a !== void 0 ? _a : 0;
            const subterm = Utils_2.fun(f, ...Utils_1.pop(stack, ar).filter(Utils_2.isSomething));
            stack.push(subterm);
        }
    }
    return stack[0];
};
const buildMatcher = (trs) => {
    const rules = [...trs.values()].flat();
    const patterns = rules.map(Utils_2.lhs);
    const patternsStr = patterns.map(exports.stringify);
    const M = new Set(patternsStr);
    const arities = exports.collectArities(patterns);
    const M_ = exports.closure(M, arities);
    const matcher = new StringClosureMatcher_1.StringClosureMatcher();
    const symbols = [...arities.keys(), exports.ω];
    const splitSymbols = genSymbolSplitter(symbols);
    for (const pat_ of M_) {
        const letters = splitSymbols(pat_);
        if (letters) {
            const pat = exports.unstringify(pat_, splitSymbols, arities);
            if (Utils_2.isNothing(pat))
                continue;
            const P_s = [];
            for (const [pattern, rule] of Utils_2.zip(patterns, rules)) {
                if (Unification_1.matches(pat, pattern)) {
                    P_s.push(rule);
                }
            }
            matcher.insert(letters, P_s);
        }
    }
    return (query) => {
        return matcher.lookupTerm(query, arities);
    };
};
exports.buildMatcher = buildMatcher;
