"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ImperativeLangTranslator = void 0;
const DecisionTree_1 = require("../Compiler/DecisionTrees/DecisionTree");
const DecisionTreeCompiler_1 = require("../Compiler/DecisionTrees/DecisionTreeCompiler");
const DecisionTreeTranslator_1 = require("../Compiler/DecisionTrees/DecisionTreeTranslator");
const Utils_1 = require("../Compiler/Utils");
const Types_1 = require("../Parser/Types");
const Utils_2 = require("../Parser/Utils");
const Translator_1 = require("./Translator");
function isFunCall(val) {
    return typeof val === 'object' && Types_1.dictHas(val, 'funName');
}
// a translator for imperative programming languages like c
class ImperativeLangTranslator extends DecisionTreeTranslator_1.DecisionTreeTranslator {
    constructor() {
        super(...arguments);
        this.compileDecisionTree = (name, dt, varNames) => {
            let hasTailRecursiveCall = false;
            const translate = (tree) => {
                switch (tree.type) {
                    case 'fail':
                        // pattern matching failed
                        // return a free instance of the term with its arguments
                        return {
                            type: 'return',
                            value: this.astOfTerm(Utils_1.fun(name, ...varNames))
                        };
                    case 'leaf':
                        const ret = this.callOccurence(tree.action, varNames);
                        // check if this is a tail recursive call
                        if (isFunCall(ret) && ret.funName === name) {
                            // create temporary variables to store updated arguments
                            // since they might require the original arguments values
                            // which should therefore not be updated right away
                            const newArgs = ret.args.map(a => this.translateExtendedTerm(a));
                            const updateArgs = Utils_2.map(Utils_1.zip(varNames, newArgs), ([v, newVal]) => ({
                                type: 'const_decl',
                                name: `${v}_`,
                                value: newVal
                            }));
                            // copy the updated arguments back
                            const copyArgs = varNames.map(v => ({
                                type: 'assign',
                                lhs: v,
                                rhs: `${v}_`
                            }));
                            hasTailRecursiveCall = true;
                            return {
                                type: 'block',
                                statements: [
                                    ...updateArgs,
                                    ...copyArgs,
                                    { type: 'continue' }
                                ]
                            };
                        }
                        else {
                            return {
                                type: 'return',
                                value: this.astOfExtendedTerm(ret)
                            };
                        }
                    case 'switch':
                        const cases = [];
                        for (const [ctor, A] of tree.tests) {
                            if (ctor === DecisionTreeCompiler_1._) {
                                cases.push({
                                    type: 'default',
                                    body: translate(A)
                                });
                            }
                            else {
                                cases.push({
                                    type: 'case',
                                    test: Translator_1.isNat(ctor) ?
                                        { type: 'nat', value: parseInt(ctor) } :
                                        { type: 'string', value: ctor },
                                    body: translate(A)
                                });
                            }
                        }
                        const occ = this.astOfOccurence(tree.occurence, varNames);
                        return {
                            type: 'switch',
                            testedValue: {
                                type: 'term_name',
                                term: occ
                            },
                            cases
                        };
                }
            };
            const tree = translate(dt);
            // if the function is tail recursive, wrap the body inside a while loop
            const body = hasTailRecursiveCall ?
                { type: 'infinite_loop', body: tree } :
                tree;
            return {
                type: 'function_decl',
                name,
                argNames: varNames,
                body
            };
        };
    }
    declareNullary(varName, symb) {
        return this.translateAst({
            type: 'const_decl',
            name: varName,
            value: this.translateTerm({ name: symb, args: [] }, false)
        });
    }
    translateTerm(term, useNullaryVars = false) {
        if (Utils_1.isVar(term)) {
            return this.translateAst({
                type: 'var',
                name: term
            });
        }
        if (useNullaryVars && term.args.length === 0) {
            if (Translator_1.isNat(term.name)) {
                return this.translateAst({
                    type: 'nat',
                    value: parseInt(term.name)
                });
            }
            return this.translateAst({
                type: 'var',
                name: Translator_1.nullaryVarName(term.name)
            });
        }
        return this.translateAst({
            type: 'fun',
            name: term.name,
            args: term.args.map(a => this.astOfTerm(a))
        });
    }
    accessSubterm(parent, childIndex) {
        return this.translateAst({
            type: 'subterm_access',
            parent: { type: 'raw', value: parent },
            childIndex
        });
    }
    translateExtendedTerm(t) {
        if (typeof t === 'string')
            return t;
        if (isFunCall(t)) {
            return this.translateAst({
                type: 'function_call',
                funName: t.funName,
                args: t.args.map(s => this.astOfExtendedTerm(s))
            });
        }
        return this.translateTerm(t, true);
    }
    translateOccurence(occ, varNames) {
        if (DecisionTree_1.isOccurence(occ)) {
            if (occ.pos.length === 0)
                return varNames[occ.index];
            return occ.pos.reduce((p, i) => this.accessSubterm(p, i), varNames[occ.index]);
        }
        if (occ.args.length === 0)
            return occ.name;
        return `${occ.name} (${occ.args.map(o => this.translateOccurence(o, varNames)).join(', ')})`;
    }
    callOccurence(occ, varNames) {
        if (DecisionTree_1.isOccurence(occ))
            return this.translateOccurence(occ, varNames);
        // if this is a free constructor simply output a term
        if (!this.isDefined(occ.name)) {
            return this.translateTerm(Utils_1.fun(occ.name, ...occ.args.map(t => this.translateExtendedTerm(this.callOccurence(t, varNames)))), true);
        }
        // otherwise call the corresponding js function
        const args = occ.args.map(t => this.translateExtendedTerm(this.callOccurence(t, varNames)));
        return { funName: occ.name, args };
    }
    astOfTerm(t) {
        if (Utils_1.isVar(t))
            return { type: 'var', name: t };
        return {
            type: 'fun',
            name: t.name,
            args: t.args.map(s => this.astOfTerm(s))
        };
    }
    astOfExtendedTerm(t) {
        if (isFunCall(t)) {
            return {
                type: 'function_call',
                funName: t.funName,
                args: t.args.map(s => this.astOfExtendedTerm(s))
            };
        }
        return this.astOfTerm(t);
    }
    astOfOccurence(occ, varNames) {
        return { type: 'raw', value: this.translateOccurence(occ, varNames) };
    }
    translateDecisionTree(name, dt, varNames) {
        return this.translateAst(this.compileDecisionTree(name, dt, varNames));
    }
    callTerm(term) {
        if (Utils_1.isVar(term))
            return term;
        // if this is a free constructor simply output a term
        if (!this.isDefined(term.name)) {
            return this.translateTerm(Utils_1.fun(term.name, ...term.args.map(t => this.callTerm(t))));
        }
        // otherwise call the corresponding js function
        const args = term.args.map(t => ({ type: 'raw', value: this.callTerm(t) }));
        return this.translateAst({
            type: 'function_call',
            funName: term.name,
            args: args
        });
    }
}
exports.ImperativeLangTranslator = ImperativeLangTranslator;
