"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Translator = exports.nullaryVarName = exports.nullarySymbolsPrefix = exports.isNat = exports.symbolMap = void 0;
const Utils_1 = require("../Compiler/Utils");
const Parser_1 = require("../Parser/Parser");
const Types_1 = require("../Parser/Types");
const Utils_2 = require("../Parser/Utils");
const Lazify_1 = require("../Compiler/Passes/Lazify");
exports.symbolMap = {
    '.': '_dot_',
    '-': '_minus_',
    '~': '_tilde_',
    '+': '_plus_',
    '*': '_star_',
    '&': '_ampersand_',
    '|': '_pipe_',
    '/': '_slash_',
    '\\': '_backslash_',
    '^': '_caret_',
    '%': '_percent_',
    '°': '_num_',
    '$': '_dollar_',
    '@': '_at_',
    '#': '_hash_',
    ';': '_semicolon_',
    ':': '_colon_',
    '_': '_',
    '=': '_eq_',
    "'": '_prime_',
    '>': '_gtr_',
    '<': '_lss_',
    '!': '_exclamation_mark_'
};
const isNat = (str) => natRegex.test(str);
exports.isNat = isNat;
exports.nullarySymbolsPrefix = 'symb_';
const natRegex = /^[0-9]+$/;
const nullaryVarName = (f) => {
    return `${exports.nullarySymbolsPrefix}${Utils_2.mapString(f, c => { var _a; return (_a = exports.symbolMap[c]) !== null && _a !== void 0 ? _a : c; })}`;
};
exports.nullaryVarName = nullaryVarName;
class Translator {
    constructor(trs, externals, reservedKeywords = new Set()) {
        this.header = [];
        this.trs = trs;
        this.externals = externals;
        this.reservedKeywords = reservedKeywords;
        this.definedSymbols = new Set([...trs.keys(), ...Object.keys(externals).map(s => `@${s}`)]
            .map(f => this.withoutSpecialChars(f)));
        this.init();
        this.declareNullarySymbols();
    }
    // eslint-disable-next-line @typescript-eslint/no-empty-function
    init() { }
    declareNullarySymbols() {
        this.nullaries = new Set();
        if (['equ', 'gtr', 'geq', 'lss', 'leq'].some(f => Types_1.dictHas(this.externals, f))) {
            this.nullaries.add('True');
            this.nullaries.add('False');
        }
        const symbs = Lazify_1.collectTRSArities(this.trs);
        for (const [symb, ar] of symbs) {
            if (ar === 0 && !exports.isNat(symb)) {
                this.nullaries.add(symb);
            }
        }
        const decl = Utils_2.setMap(this.nullaries, f => this.declareNullary(exports.nullaryVarName(f), f));
        this.header.push(...decl);
    }
    setReservedKeywords(kw) {
        this.reservedKeywords = kw;
    }
    withSpecialChars(str) {
        return Utils_1.replaceSubstrings(str, exports.symbolMap);
    }
    withPrefix(name) {
        return `grf_${name}`;
    }
    withoutPrefix(str) {
        return str.replace(new RegExp('grf_', 'g'), '');
    }
    withoutSpecialChars(name) {
        const noSymbols = Utils_2.mapString(name, c => { var _a; return (_a = exports.symbolMap[c]) !== null && _a !== void 0 ? _a : c; });
        return this.withPrefix(noSymbols);
    }
    parseRenamedTerm(term) {
        return Parser_1.parseTerm(this.withoutPrefix(this.withSpecialChars(term)));
    }
    isDefined(f) {
        return this.trs.has(f) || this.definedSymbols.has(f) ||
            Types_1.dictHas(this.externals, f.substr(1));
    }
    renameVar(v) {
        if (this.reservedKeywords.has(v)) {
            return `_${v}`;
        }
        return v;
    }
    renameTerm(term) {
        if (Utils_1.isVar(term))
            return this.renameVar(term);
        if (!this.isDefined(term.name)) {
            return Utils_1.fun(term.name, ...term.args.map(t => this.renameTerm(t)));
        }
        return this.renameFun(term);
    }
    renameFun(term) {
        return {
            name: this.withoutSpecialChars(term.name),
            args: term.args.map(t => this.renameTerm(t)),
        };
    }
    renameRule([lhs, rhs]) {
        return [this.renameFun(lhs), this.renameTerm(rhs)];
    }
    generateExternals() {
        const exts = [];
        for (const [name, gen] of Object.entries(this.externals)) {
            exts.push(gen(this.withoutSpecialChars("@" + name)));
        }
        return exts.join("\n");
    }
    generateRules() {
        const res = [];
        for (const [name, rules] of this.trs) {
            res.push(this.translateRules(this.withoutSpecialChars(name), rules.map(r => this.renameRule(r))));
        }
        return res.join("\n");
    }
    translate() {
        return [
            this.header.join("\n"),
            this.generateExternals(),
            this.generateRules(),
        ].join("\n");
    }
}
exports.Translator = Translator;
